// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

#ifndef P_GRAPH_H_
#define P_GRAPH_H_

/*###########################################################################*/
#include "RVector.h"
#include "TTdag.h"

class ostream;
class ResourceInstance;
class Resource;

// --------------------------------------------------------------------------
class ResourceAttrib {
	friend class Resource;
	friend class PGraph;
	friend double com_possible(RString aname, double t1, double t2,
							   RVector<ResourceInstance> &coms,
							   RVector<Resource> &com_types,
							   double arc_fill_factor);
private:
	ResourceAttrib	(RString nm, double v) {name_ = nm; value_ = v;};
	RString	name_;
	double	value_;
};

class ResourceUse {
	friend class ResourceInstance;
	friend class PGraph;
	friend ResourceUse *NextToStart(double t, RVector<ResourceUse>& ruv);
	friend double com_possible(RString aname, double t1, double t2,
							   RVector<ResourceInstance> &coms,
							   RVector<Resource> &com_types,
							   double arc_fill_factor);
private:
	ResourceUse	(RString nm, double start, double end)
					{name_ = nm; ts_ = start; te_ = end;};
	RString	name_;
	double	ts_, te_;
	void print_to(ostream &os) const;
	friend ostream &operator<<(ostream &os, /**/ const ResourceUse &ru)
	  {ru.print_to(os); return os;}
};

class Resource {
	friend class ResourceInstance;
	friend class PGraph;
private:
	RString					name_;
	RVector<ResourceAttrib>	attrib_;
	RVector<double>			exec_time_;	// for PE resources only

	void print_to(ostream &os) const;
	friend ostream &operator<<(ostream &os, /**/ const Resource &r)
	  {r.print_to(os); return os;}
public:
	ResourceAttrib& GetAttrib(RString s);
	void SetAttrib(RString s, double v);
};

class ResourceInstance {
	friend class PGraph;
	friend double com_possible(RString aname, double t1, double t2,
							   RVector<ResourceInstance> &coms,
							   RVector<Resource> &com_types,
							   double arc_fill_factor);
private:
	ResourceInstance(RString nm, unsigned r, double tf)
					{name_ = nm; resource_indx_ = r; time_frame_ = tf;};

	RString					name_;
	unsigned				resource_indx_;
	double					time_frame_;
	RVector<ResourceUse>	ru_;

	void print_to(ostream &os) const;
	friend ostream &operator<<(ostream &os, /**/ const ResourceInstance &ri)
	  {ri.print_to(os); return os;}
};

/*###########################################################################*/
class PGraph {
public:
	PGraph(	ostream &eps_out,
			int	   num_task_graphs,
			int	   avg_tasks_per_pe,
			double avg_task_time,	/* PE STUFF */
			double mul_task_time,
			double task_slack,
			double task_round,
			int    num_pe_types,
			int    num_pe_soln,
			int    num_com_types,	/* COM STUFF */
			int    num_com_soln,
			double arc_fill_factor );
	void print_to(ostream &os) const;
	void print_to_ps(ostream &os) const;
	double h_period() const {return h_period_;}
	friend ostream &operator<<(ostream &os, /**/ const PGraph &pg)
	  {pg.print_to(os); return os;}

private:
	RVector<Resource>			pe_;
	RVector<ResourceInstance>	pe_soln_;
	RVector<Resource>			com_;
	RVector<ResourceInstance>	com_soln_;
	RVector<int>				arc_data_size_;
	RVector<TTdag>				dag_; 
	double						cost_of_soln_;
	double h_period_;		// only non-periodic now, keep for future
};

/*###########################################################################*/
#endif

