// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef R_PAIR_H_
#define R_PAIR_H_

/*###########################################################################*/
#include <utility>
#include <functional>
#include <iostream>
#include "RStd.h"

/*###########################################################################*/
template <typename T> struct HPair : public pair<T, T> {
	typedef pair<T, T> super;
	typedef unsigned size_type;

	HPair() {}
	HPair(const T &a, const T &b) : super(a, b) {}
	template <class U1, class U2>
	  HPair(const pair<U1, U2>& p) : super(p) {}
	T &operator[](size_type x);
	const T &operator[](size_type x) const;
};

/*===========================================================================*/
template <typename T1, typename T2>
ostream &operator<<(ostream &os, const pair<T1, T2> &p)
  {return os << "[" << p.first << " " << p.second << "]";}

/*===========================================================================*/
template <typename Pair>
struct equal_to_1st : public binary_function<Pair, Pair, bool> {
    bool operator()(const Pair &x, const Pair &y) const
	  {return x.first == y.first; }
}; 

template <typename Pair>
struct less_1st : public binary_function<Pair, Pair, bool> {
    bool operator()(const Pair &x, const Pair &y) const
	  {return x.first < y.first; }
}; 

/*===========================================================================*/
template <typename Pair>
struct equal_to_2nd : public binary_function<Pair, Pair, bool> {
    bool operator()(const Pair &x, const Pair &y) const
	  {return x.second == y.second; }
}; 

template <typename Pair>
struct less_2nd : public binary_function<Pair, Pair, bool> {
    bool operator()(const Pair &x, const Pair &y) const
	  {return x.second < y.second; }
}; 

/*###########################################################################*/
template <typename T1, typename T2, typename T3>
struct Triple {
	typedef T1 first_type;
	typedef T2 second_type;
	typedef T3 third_type;

	T1 first;
	T2 second;
	T3 third;

	Triple() : first(T1()), second(T2()), third(T3()) {}
	Triple(const T1 &x, const T2 &y, const T3 &z) :
	  first(x), second(y), third(z) {}

	template <typename U1, typename U2, typename U3>
	  Triple(const Triple<U1, U2, U3> &p) :
	  first(p.first), second(p.second), third(p.third) {}
};

/*===========================================================================*/
template <typename T1, typename T2, typename T3>
ostream &operator<<(ostream &os, const Triple<T1, T2, T3> &t)
  {return os << "[" << t.first << " " << t.second << " " << t.third << "]";}

/*===========================================================================*/
template <typename T>
struct HTriple : public Triple<T, T, T> {
	typedef Triple<T, T, T> super;
	typedef unsigned size_type;

	HTriple() {}
	HTriple(const T &x, const T &y, const T &z) : super(x, y, z) {}
	template <typename U1, typename U2, typename U3>
	  HTriple(const Triple<U1, U2, U3> &p) : super(p) {}
	T &operator[](size_type x);
	const T &operator[](size_type x) const;
};

/*===========================================================================*/
template <typename T1, typename T2, typename T3>
  inline bool operator==(const Triple<T1, T2, T3> &x,
  const Triple<T1, T2, T3> &y);

template <typename T1, typename T2, typename T3>
  inline bool operator<(const Triple<T1, T2, T3> &x,
  const Triple<T1, T2, T3> &y);

/*===========================================================================*/
template <typename T1, typename T2, typename T3>
inline Triple<T1, T2, T3> make_triple(const T1 &x, const T2 &y, const T3 &z)
  {return Triple<T1, T2, T3>(x, y, z);}

/*===========================================================================*/
template <typename Triple>
struct select3rd :
public unary_function<Triple, typename Triple::third_type> {
	const typename Triple::third_type &operator()(const Triple &x) const
	  {return x.third;}
};

template <typename Triple>
struct equal_to_3rd : public binary_function<Triple, Triple, bool> {
    bool operator()(const Triple &x, const Triple &y) const
	  {return x.third == y.third; }
}; 

template <typename Triple>
struct less_3rd : public binary_function<Triple, Triple, bool> {
    bool operator()(const Triple &x, const Triple &y) const
	  {return x.third < y.third; }
}; 

/*###########################################################################*/
#include "RPair.cct"
#endif

