// Copyright 2000 by Robert Dick.
// All rights reserved.

#include <cstdlib>
#include <iostream>
#include <new>
#include "RStd.h"

/*###########################################################################*/
ostream cerrhex(cerr.rdbuf());
int cerr_hex_init() {cerrhex << hex; return 0;}
static int L_cerr_hex_init = cerr_hex_init();

ostream couthex(cout.rdbuf());
int cout_hex_init() {couthex << hex; return 0;}
static int L_cout_hex_init = cout_hex_init();

/*===========================================================================*/
void rexit(const char *s) {
	cout << s;
	exit(EXIT_SUCCESS);
}

/*===========================================================================*/
void rpos_(const char *file, int line) {
	cerr << "File: " << file << " Line: " << line << endl;
}

/*===========================================================================*/
void rabort_(const char *file, int line) {
	cerr << "Program aborted.\n";
	rpos_(file, line);
	abort();
}

/*===========================================================================*/
void rassert_(bool x, const char *file, int line) {
	if (! x) {
		cerr << "Assertion failure.\n";
		rpos_(file, line);
		abort();
	}
}

/*###########################################################################*/
#ifdef ROB_DEBUG

int rob_debug_ = 1;

int MemChecker::out_mem_ = 0;
bool MemChecker::watching_ = true;

/*===========================================================================*/
void MemChecker::inc() {
	out_mem_++;
}

/*===========================================================================*/
void MemChecker::dec() {
	out_mem_--;
	if (watching_ && out_mem_ < 0) {
		cerr << "***** negative outstanding object" <<
		  (out_mem_ == 1 ? "\n" : "s\n");
		abort();
	}
}

/*===========================================================================*/
void MemChecker::set_mem(int mem) {
	out_mem_ = mem;
}

/*===========================================================================*/
int MemChecker::out() {
	return out_mem_;
}

/*===========================================================================*/
void MemChecker::set_watch(bool watch) {
	watching_ = watch;
}

/*===========================================================================*/
void MemChecker::check(ostream &os) {
	if (out_mem_) {
		os << "*** " << out_mem_ << " allocation" <<
		  (out_mem_ == 1 ? " out\n" : "s out\n");
	}
}

/*###########################################################################*/
static const size_t ALIGN_PAD = sizeof(long double) / sizeof(size_t);

/*===========================================================================*/
void *operator new(size_t bytes) {
	if (! bytes)
		throw bad_alloc();

	MemChecker::inc();
	size_t *mem = (size_t *)malloc(bytes + ALIGN_PAD * sizeof(size_t));
	*mem = bytes;
	return mem + ALIGN_PAD;
}

/*===========================================================================*/
void operator delete(void *data) {
	if (! data)
		return;

	size_t *mem = (size_t *)data;
	size_t mem_size = *(mem - ALIGN_PAD);

	Rassert(mem_size);

	unsigned char *mem_char = (unsigned char *)data;

	for (size_t cnt = 0; cnt < mem_size; cnt++)
		mem_char[cnt] = 0;

	*(mem - ALIGN_PAD) = 0;
	free(mem - ALIGN_PAD);
	MemChecker::dec();
}

/*===========================================================================*/
void *operator new[](size_t bytes) {
	if (! bytes)
		throw bad_alloc();

	MemChecker::inc();
	size_t *mem = (size_t *)malloc(bytes + ALIGN_PAD * sizeof(size_t));
	*mem = bytes;
	return mem + ALIGN_PAD;
}

/*===========================================================================*/
void operator delete[](void *data) {
	if (! data)
		return;

	size_t *mem = (size_t *)data;
	size_t mem_size = *(mem - ALIGN_PAD);

	Rassert(mem_size);

	unsigned char *mem_char = (unsigned char *)data;

	for (size_t cnt = 0; cnt < mem_size; cnt++)
		mem_char[cnt] = 0;

	*(mem - ALIGN_PAD) = 0;
	free(mem - ALIGN_PAD);
	MemChecker::dec();
}
#endif
