// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef R_STD_H_
#define R_STD_H_

// Define ROB_DEBUG to turn on debugging.

/*###########################################################################*/
class ostream;
extern ostream cerrhex;
extern ostream couthex;

#undef PI
const double PI = 3.14159265358979323846;

/*===========================================================================*/
// Utility

#define MAP(x, end_x) \
	for (unsigned x = 0; (x) < (unsigned)(end_x); (x)++)

#define MAP2(x, end_x, y, end_y) \
	for (unsigned x = 0; (x) < (unsigned)(end_x); (x)++) \
		for (unsigned y = 0; (y) < (unsigned)(end_y); (y)++)

#define MAP3(x, end_x, y, end_y, z, end_z) \
	for (unsigned x = 0; (x) < (unsigned)(end_x); (x)++) \
		for (unsigned y = 0; (y) < (unsigned)(end_y); (y)++) \
			for (unsigned z = 0; (z) < (unsigned)(end_z); (z)++)

inline bool xor(bool a, bool b) {return a != b;}

/*===========================================================================*/
// Debugging
void rexit(const char *s = "");
void rpos_(const char *file, int line);
void rabort_(const char *file, int line);
void rassert_(bool x, const char *file, int line);

/* The behavior of Rdump() and Rdumphex() are not defined until all
global constructors have been executed. */
#define Rdump(x) (cerr << #x << ": " << (x) << "\n")
#define Rdumphex(x) (cerrhex << #x << ": 0x" << (x) << "\n")
#define Rpos() rpos_(__FILE__, __LINE__)
#define Rabort() rabort_(__FILE__, __LINE__)
#define Rassert(x) rassert_((x), __FILE__, __LINE__)

#ifndef ROB_DEBUG
#define RPOS() do {} while (0)
#define RDEBUG(x) do {} while (0)
#define RDEBUG_START do {
#define RDEBUG_END } while (0);
#define RASSERT(x) do {} while (0)
#define RLEVEL(x) do {} while (0)
#define MEM_CHECK(x) do {} while (0)
#else

extern int rob_debug_;
#define RPOS() if (rob_debug_) rpos_(__FILE__, __LINE__)
#define RDEBUG(x) if (rob_debug_) (x)
#define RDEBUG_START if (rob_debug_) {
#define RDEBUG_END }
#define RASSERT(x) if (rob_debug_) rassert_((x), __FILE__, __LINE__)
#define RDUMP(x) if (rob_debug_) cerr << #x << ": " << (x) << "\n"
#define RDUMPHEX(x) if (rob_debug_) cerrhex << #x << ": 0x" << (x) << "\n"
#define RLEVEL(x) (rob_debug_ = (x))
#define MEM_CHECK(x) MemChecker::check(x)

/*===========================================================================*/
class MemChecker {
/* Redefines new and delete to track outstanding heap allocations.
Relies on the last object on the link line being the first translation
unit in which static and global initializations are done.
This is just for debugging. */

public:
	static void start();
	static void inc();
	static void dec();
	static void set_mem(int mem);
	static int out();
	static void set_watch(bool watch);
	static void check(class ostream &os);

private:
		static int out_mem_;
		static bool watching_;
};

#endif

/*###########################################################################*/
#endif
