// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

//
//    File: listman.cc
//
//	Author: David L Rhodes
// Version: v2.0  --  December 29, 1997
//
// Purpose: Implement (P)SubListman member functions
//

#include "listman.h"

// -------------------------------------------------------
// ---SubListman------------------------------------------
// -------------------------------------------------------

// -------------------------------------------------------
SubListman::SubListman()
{
	// create an empty list
	root = new SubListObj;
	root->next = root->prev = root;
	root->data = 0;
	n_onlist = 0;
}
// --------------------------------------------------------
SubListman::~SubListman()
{
	// destroy everything
	Empty();
	delete root;
}
// --------------------------------------------------------
void SubListman::Empty()
{
	// destroy everything
	SubListObj *p = (SubListObj *)root->next;
	while( p != root ) {
		SubListObj *nxt = (SubListObj *)p->next;
		delete p;
		p = nxt;
	}

	root->next = root->prev = root;
	n_onlist = 0;
}
// --------------------------------------------------------
int SubListman::OnList(SubListObj *obj) const
{
	SubListObj *p = (SubListObj *)root->next;
	while( p != root ) {
		if( obj == p ) return 1;
		p = (SubListObj *)p->next;
	}

	return 0; // not found
}
// --------------------------------------------------------
void SubListman::Remove(SubListObj *obj)
{
	RASSERT( OnList(obj) );
	obj->next->prev = (SubListObj *)obj->prev;
	obj->prev->next = (SubListObj *)obj->next;

	delete obj;
	n_onlist--;
}
// --------------------------------------------------------
SubListObj *SubListman::operator[](int k) const
{
	// return k-th element ptr where k=0 implies first in list
	if( k >= n_onlist || k < 0 ) return 0;
	SubListObj *p = (SubListObj *)root->next;
	for(int j=1; j <= k; j++ ) {
		p = (SubListObj *)p->next;
	}

	return p;
}
// --------------------------------------------------------
/*inline*/ SubListObj *SubListman::First(void) const
{
	return ((root->next == root) ? 0 : (SubListObj *)root->next);
}
// --------------------------------------------------------
/*inline*/ SubListObj *SubListman::Last(void) const
{
	return ((root->prev == root) ? 0 : (SubListObj *)root->prev);
}
// --------------------------------------------------------
/*inline*/ SubListObj *SubListman::Next(SubListObj *curobj) const
{
	return ((curobj->next == root) ? 0 : (SubListObj *)curobj->next);
}
// --------------------------------------------------------
/*inline*/ SubListObj *SubListman::Prev(SubListObj *curobj) const
{
	return ((curobj->prev == root) ? 0 : (SubListObj *)curobj->prev);
}
// --------------------------------------------------------
SubListObj *SubListman::InsertAfter(SubListObj *listobj,void *listaddition)
{
	// if listobj == NULL then the insert is at the head
	RASSERT( !listobj || OnList(listobj) );
	SubListObj *before = (listobj == 0 ? root : listobj);
	SubListObj *after  = (SubListObj *)before->next;
	SubListObj *t = new SubListObj;
	t->next = after;
	t->prev = before;
	t->data = listaddition;
	before->next = after->prev = t;
	
	n_onlist++;
	return t;
}
// --------------------------------------------------------
SubListObj *SubListman::InsertBefore(SubListObj *listobj,void *listaddition)
{
	// if listobj == NULL then the insert is at the tail
	RASSERT( !listobj || OnList(listobj) );
	SubListObj *after  = (listobj == 0 ? root : listobj);
	SubListObj *before = (SubListObj *)after->prev;
	SubListObj *t = new SubListObj;
	t->next = after;
	t->prev = before;
	t->data = listaddition;
	before->next = after->prev = t;
	
	n_onlist++;
	return t;
}
// --------------------------------------------------------
