// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

#include <fstream.h>
#include <iomanip>
#include <strstream.h>
#include <cmath>
#include <algorithm>
#include <functional>
#include <map>
#include "RMath.h"
#include "RStd.h"
#include "RGen.h"
#include "ArgPack.h"
#include "TGraph.h"
#include "PGraph.h"
#include "DBase.h"

/*###########################################################################*/
// Returns the hyperperiod of all the graphs produced.
static double parse(ostream &data_out, ostream &eps_out, istream &opt_in);

static void parse_error(int line);

/*###########################################################################*/
int main(int argc, char *argv[]) {
#if 0
while (1) {
	long a = 0, b = 0, c = 0, d = 0, e = 0, f = 0, g = 0,
	  h = 0, i = 0, j = 0;

	double total = 0.0;

	long ticker = 0;

	while (1) {
		double alpha = RGen::gen.flat_range_d(0.0, 1.0);
		double beta = 1.0 - alpha;

		double urv_a = RGen::gen.flat_range_d(-1.0, 1.0);
		double urv_b = RGen::gen.flat_range_d(-1.0, 1.0);

		double fred = DBase::flat_merge(alpha, urv_a, urv_b);
	
		fred += 1.0;
		fred *= 5.0;

		if (fred < 1.0)
			a++;
		else if (fred < 2.0)
			b++;
		else if (fred < 3.0)
			c++;
		else if (fred < 4.0)
			d++;
		else if (fred < 5.0)
			e++;
		else if (fred < 6.0)
			f++;
		else if (fred < 7.0)
			g++;
		else if (fred < 8.0)
			h++;
		else if (fred < 9.0)
			i++;
		else
			j++;

		total += 1.0;

		if (ticker == 10000) {
			cout << setw(10) << a / total << "\t" <<
			  setw(10) << b / total << "\t" <<
			  setw(10) << c / total << "\t" <<
			  setw(10) << d / total << "\t" <<
			  setw(10) << e / total << "\t" <<
			  setw(10) << f / total << "\t" <<
			  setw(10) << g / total << "\t" <<
			  setw(10) << h / total << "\t" <<
			  setw(10) << i / total << "\t" <<
			  setw(10) << j / total << "\n";

			ticker = 0;
		}
		ticker++;
	}
}

rexit();
#endif

	ArgPack::init(argc, argv);

	ifstream opt_in(ArgPack::opt_file_name_.begin());
	if (! opt_in) {
		ArgPack::help(cout);
		rexit("Unable to open input file.\n");
	}

	ofstream eps_out(ArgPack::eps_file_name_.begin());
	if (! eps_out) {
		ArgPack::help(cout);
		rexit("Unable to open output files.\n");
	}

	ostrstream data_buf;

	data_buf.setf(ios::left);
	double h_period = parse(data_buf, eps_out, opt_in);

	ofstream data_out(ArgPack::data_file_name_.begin());
	if (! data_out) {
		ArgPack::help(cout);
		rexit("Unable to open output files.\n");
	}

	if (h_period > 0.0)
		data_out << "@HYPERPERIOD " << h_period << "\n\n";

	data_out << data_buf.str();

//	MEM_CHECK(cout);
	rexit();
}

/*===========================================================================*/
static double parse(ostream &data_out, ostream &eps_out, istream &opt_in) {
	RVector<double> h_period;

	int line = 0;
	map<RString, int> tg_offset;
	map<RString, int> pe_offset;

	tg_offset.insert(make_pair(ArgPack::tg_label_, 0));
	pe_offset.insert(make_pair(ArgPack::table_label_, 0));

	while (1) {
		RString s;
		bool wrap = true;

// Get a single command
		do {
			RString line_buf;
			if (! line_buf.append_line(opt_in)) {
				if (h_period.size())
					return lcm(h_period);
				else
					return -1.0;
			}

			line++;

			pair<RString, RString::const_iterator> first_word =
			  line_buf.token(line_buf.begin());

			if (line_buf.front() != '#' && ! first_word.first.empty()) {
				s += line_buf;

				if (s.back() == '\\')
					s.pop_back();
				else
					wrap = false;
			}
		} while (wrap);

		pair<RString, RString::iterator> tok = s.token(s.begin());
		if (tok.first.empty())
			parse_error(line);

		s.erase(s.begin(), tok.second);
		const RString &command = tok.first;

		RVector<RString> vec = s.tokenize(" \t");
		RVector<RString> top_rvec(0);
		if( vec.size() > 0 )			// something there
			top_rvec = s.tokenize(",");

		RVector2<RString> rvec(top_rvec.size());

		MAP(x, rvec.size())
			rvec[x] = top_rvec[x].tokenize();

/*-------------------------------------*/
		if (command == "tg_label") {
			if (vec.size() != 1) parse_error(line);
			ArgPack::tg_label_ = vec[0];
			tg_offset.insert(make_pair(ArgPack::tg_label_, 0));

		} else if (command == "tg_cnt") {
			if (vec.size() != 1 ||
			  (ArgPack::tg_cnt_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "task_trans_time") {
			if (vec.size() != 1 ||
			  (ArgPack::vertex_time_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "pack_schedule") {
			if ( !(vec.size() == 10 || vec.size() == 11) )
				parse_error(line);

			int		num_task_graphs	 = vec[0].convert();
			int		avg_tasks_per_pe = vec[1].convert();
			if( num_task_graphs < 1 || avg_tasks_per_pe < 1)
				parse_error(line);

			double	avg_task_time	= vec[2].convert();
			double	mul_task_time	= vec[3].convert();
			double	task_slack		= vec[4].convert();
			double	task_round		= vec[5].convert();
			int		num_pe_types	= vec[6].convert();
			int		num_pe_soln		= vec[7].convert();
			if( avg_task_time <= 0.0 ||
				avg_task_time+abs(mul_task_time) <= 0.0 ||
				task_slack < 0.0 ||
				num_pe_types < 1 ||
				num_pe_soln < 1 )
				parse_error(line);

			int		num_com_types	= vec[8].convert();
			int		num_com_soln	= vec[9].convert();
			if( num_com_types < 1 ||
				num_com_soln < 1 )
				parse_error(line);
			double	arc_fill_factor	= 1.0;
			if( vec.size() == 11 )
				arc_fill_factor	= vec[10].convert();

			// here goes...
			PGraph pg(	eps_out,
						num_task_graphs,
						avg_tasks_per_pe,
						avg_task_time,	/* PE STUFF */
						mul_task_time,
						task_slack,
						task_round,
						num_pe_types,
						num_pe_soln,
						num_com_types,	/* COM STUFF */
						num_com_soln,
						arc_fill_factor );

			data_out << pg;
			return pg.h_period();

		} else if (command == "task_cnt") {
			if (vec.size() != 2) parse_error(line);
			ArgPack::vertex_cnt_av_ = vec[0].convert();
			ArgPack::vertex_cnt_mul_ = vec[1].convert();

			if (ArgPack::vertex_cnt_av_ - abs(ArgPack::vertex_cnt_mul_) < 0)
				parse_error(line);

		} else if (command == "task_degree") {
			if (vec.size() != 2) parse_error(line);
			ArgPack::vertex_in_deg_ = vec[0].convert();
			ArgPack::vertex_out_deg_ = vec[1].convert();

			/* when using "pack_schedule" -1 indicates "no limit"
			if (ArgPack::vertex_in_deg_ < 1 || ArgPack::vertex_out_deg_ < 1)
				parse_error(line);
			*/

		} else if (command == "task_unique") {
			if (vec.size() > 1) parse_error(line);
			if (vec.size())
				ArgPack::task_unique_ = vec[0].convert();
			else
				ArgPack::task_unique_ = true;

		} else if (command == "task_type_cnt") {
			if (vec.size() != 1 ||
			  (ArgPack::task_type_cnt_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "trans_unique") {
			if (vec.size() > 1) parse_error(line);
			if (vec.size() == 1)
				ArgPack::trans_unique_ = vec[0].convert();
			else
				ArgPack::trans_unique_ = true;

		} else if (command == "trans_type_cnt") {
			if (vec.size() != 1 ||
			  (ArgPack::trans_type_cnt_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "task_attrib") {
			ArgPack::task_attrib_name_.clear();
			ArgPack::task_attrib_av_.clear();
			ArgPack::task_attrib_mul_.clear();
			ArgPack::task_attrib_round_.clear();

			MAP(x, rvec.size()) {
				if (rvec[x].size() < 3 || rvec[x].size() > 4)
					parse_error(line);

				ArgPack::task_attrib_name_.push_back(rvec[x][0]);
				ArgPack::task_attrib_av_.push_back(rvec[x][1].convert());
				ArgPack::task_attrib_mul_.push_back(rvec[x][2].convert());
	
				if (rvec[x].size() >= 4)
					ArgPack::task_attrib_round_.push_back(rvec[x][3].convert());
				else
					ArgPack::task_attrib_round_.push_back(1);
			}

		} else if (command == "period_laxity") {
			if (vec.size() != 1 ||
			  (ArgPack::p_laxity_ = vec[0].convert()) <= 0.0)
				parse_error(line);

		} else if (command == "period_g_deadline") {
			if (vec.size() > 1) parse_error(line);
			if (vec.size() == 1)
				ArgPack::p_greater_deadline_ = vec[0].convert();
			else
				ArgPack::p_greater_deadline_ = true;

		} else if (command == "prob_hard_deadline") {
			if (vec.size() > 1) parse_error(line);
			if (vec.size() == 1)
				ArgPack::prob_hard_deadline_ = vec[0].convert();

			if (ArgPack::prob_hard_deadline_ < 0.0 ||
				ArgPack::prob_hard_deadline_ > 1.0)
			{
				parse_error(line);
			}

		} else if (command == "soft_deadline_mul") {
			if (vec.size() > 1) parse_error(line);
			if (vec.size() == 1)
				ArgPack::soft_deadline_mul_ = vec[0].convert();

			if (ArgPack::soft_deadline_mul_ < 0.0)
				parse_error(line);
			
		} else if (command == "period_mul") {
			ArgPack::period_mul_.clear();
			MAP(x, rvec.size()) {
				if (rvec[x].size() != 1)
					parse_error(line);
				ArgPack::period_mul_.push_back(rvec[x][0].convert());

				if (ArgPack::period_mul_.back() < 1)
					parse_error(line);
			}

		} else if (command == "entries_per_type") {
			if (vec.size() != 2) parse_error(line);
			ArgPack::entries_per_type_av_ = vec[0].convert();
			ArgPack::entries_per_type_mul_ = vec[1].convert();

			if (rint(ArgPack::entries_per_type_av_ -
				abs(ArgPack::entries_per_type_mul_)) < 0.0)
			{
				parse_error(line);
			}


		} else if (command == "urv_merge") {
			if (vec.size() != 1) parse_error(line);
			ArgPack::urv_merge_ = vec[0].convert();

		} else if (command == "table_label") {
			if (vec.size() != 1) parse_error(line);
			ArgPack::table_label_ = vec[0];
			pe_offset.insert(make_pair(ArgPack::table_label_, 0));

		} else if (command == "table_cnt") {
			if (vec.size() != 1 ||
			  (ArgPack::table_cnt_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "type_table_ratio") {
			if (vec.size() != 1 ||
			  (ArgPack::table_type_ratio_ = vec[0].convert()) < 0)
				parse_error(line);

		} else if (command == "table_attrib") {
			ArgPack::table_name_.clear();
			ArgPack::table_av_.clear();
			ArgPack::table_mul_.clear();
			ArgPack::table_jitter_.clear();
			ArgPack::table_round_.clear();

			MAP(x, rvec.size()) {
				if (rvec[x].size() < 3 || rvec[x].size() > 5)
					parse_error(line);

				ArgPack::table_name_.push_back(rvec[x][0]);
				ArgPack::table_av_.push_back(rvec[x][1].convert());
				ArgPack::table_mul_.push_back(rvec[x][2].convert());
	
				if (rvec[x].size() >= 4)
					ArgPack::table_jitter_.push_back(rvec[x][3].convert());
				else
					ArgPack::table_jitter_.push_back(0.5);

				if (rvec[x].size() == 5)
					ArgPack::table_round_.push_back(rvec[x][4].convert());
				else
					ArgPack::table_round_.push_back(1.0);
			}

		} else if (command == "type_attrib") {
			ArgPack::type_name_.clear();
			ArgPack::type_av_.clear();
			ArgPack::type_mul_.clear();
			ArgPack::type_jitter_.clear();
			ArgPack::type_round_.clear();

			MAP(x, rvec.size()) {
				if (rvec[x].size() < 3 || rvec[x].size() > 5)
					parse_error(line);

				ArgPack::type_name_.push_back(rvec[x][0]);
				ArgPack::type_av_.push_back(rvec[x][1].convert());
				ArgPack::type_mul_.push_back(rvec[x][2].convert());

				if (rvec[x].size() >= 4)
					ArgPack::type_jitter_.push_back(rvec[x][3].convert());
				else
					ArgPack::type_jitter_.push_back(0.5);

				if (rvec[x].size() == 5)
					ArgPack::type_round_.push_back(rvec[x][4].convert());
				else
					ArgPack::type_round_.push_back(1.0);					
			}

		} else if (command == "seed") {
			if (vec.size() != 1) parse_error(line);
			ArgPack::seed_ = vec[0].convert();

		} else if (command == "misc_type_cnt") {
			if (vec.size() != 1) parse_error(line);
			ArgPack::misc_type_cnt_ = vec[0].convert();

		} else if (command == "note_write") {
			MAP(x,vec.size())
				data_out << vec[x] << " ";

			data_out << endl;

		} else if (command == "eps_write") {
			RGen::gen.set_seed(ArgPack::seed_);
			TGraph tg(tg_offset[ArgPack::tg_label_]);
			tg.print_to_ps(eps_out);

		} else if (command == "tg_write") {
			RGen::gen.set_seed(ArgPack::seed_);
			TGraph tg(tg_offset[ArgPack::tg_label_]);
			tg_offset[ArgPack::tg_label_] += ArgPack::tg_cnt_;
			data_out << tg;
			h_period.push_back(tg.h_period());

		} else if (command == "pe_write") {
			RGen::gen.set_seed(ArgPack::seed_);
			DBase db(pe_offset[ArgPack::table_label_], ArgPack::task_type_cnt_);
			pe_offset[ArgPack::table_label_] += ArgPack::table_cnt_;
			data_out << db;

		} else if (command == "trans_write") {
			RGen::gen.set_seed(ArgPack::seed_);

// ugly hack
			double old_av = ArgPack::entries_per_type_av_;
			ArgPack::entries_per_type_av_ = 1.0;

			double old_mul = ArgPack::entries_per_type_mul_;
			ArgPack::entries_per_type_mul_ = 0.0;

			DBase db(pe_offset[ArgPack::table_label_], ArgPack::trans_type_cnt_);
			pe_offset[ArgPack::table_label_] += ArgPack::table_cnt_;
			db.trans_print_to(data_out);

			ArgPack::entries_per_type_av_ = old_av;
			ArgPack::entries_per_type_mul_ = old_mul;
		} else if (command == "misc_write") {
			RGen::gen.set_seed(ArgPack::seed_);
			DBase db(pe_offset[ArgPack::table_label_], ArgPack::misc_type_cnt_);
			pe_offset[ArgPack::table_label_] += ArgPack::table_cnt_;
			data_out << db;

		} else if (command == "opt_write") {
			ifstream is(ArgPack::opt_file_name_.begin());
			if (! is) parse_error(line);
	
			strstreambuf tmp;
			do {
				if (! is.get(tmp))
					parse_error(line);

				data_out << "# " << tmp.str();
				tmp.freeze(0);
			} while (is);

		} else {
			parse_error(line);
		}
	}
}

/*===========================================================================*/
static void parse_error(int line) {
	ArgPack::help(cout);
	cout << "Parse error on line " << line << ".\n";
	rexit();
}
